from util import *

log = logging.getLogger("rosemary.sqlfilter")

class SqlFilter(object):
    pass

class SqlOpFilter(SqlFilter):
    def __init__(self, operator, *ops):
        self.operator = operator
        self.valid_operators = ops
        self.check_operator()

    def check_operator(self):
        if self.valid_operators and not self.operator in self.valid_operators:
            raise Exception("Operator '%s' is not in the valid set %s" % 
                            (self.operator, self.valid_operators))

class SqlComparisonFilter(SqlOpFilter):
    '''
    Emit a comparison filter for a SQL query.
    
    The filter will be of the form 'this op that'.

    If a cursor generation callback is provided and
    'mogrify' indicates that any of the three arguments
    may need special quoting, the psycopg2.cursor.mogrify()
    routine will be used to produce a filter with
    correctly quoted values.

    If 'gen_cursor' is None or 'mogrify' does not specify
    any values to be quoted, simple concatenation will be done.

    gen_cursor -- callback to retreive a pyscopg2
    cursor, invoked as gen_cursor()

    mogrify -- which values need to be quoted.  This is an
    ordered tuple of booleans corresponding to 'this', 'operator', 
    and 'that'.  The default is False, False, True since 'this'
    and 'operator' typically come directly from code but 'that'
    may originate outside of the program.
    '''
    def __init__(self, this, that, 
                 operator="=", 
                 gen_cursor=None, 
                 mogrify=(False,False,True),
                 valid_operators=("<",">","=","<=",">=")):
        
        super(SqlComparisonFilter, self).__init__(operator, *valid_operators)

        assert isinstance(operator, str)
        assert len(mogrify) == 3

        self.this = getattr(this, "identifier", this)
        self.that = getattr(that, "identifier", that)
        self.gen_cursor = gen_cursor
        self.mogrify = mogrify
    
    def _mogrify(self):
        pattern = []
        vals = []
        for elem in zip((self.this, self.operator, self.that), self.mogrify):
            if elem[1]:
                # We want to mogrify this
                pattern.append("%s")
                vals.append(elem[0])
            else:
                # make this a literal
                pattern.append(elem[0])
        pattern = " ".join(pattern)
        cursor = self.gen_cursor()
        return cursor.mogrify(pattern, vals)

    def emit(self):
        self.check_operator()
        if self.gen_cursor and True in self.mogrify:
            return self._mogrify()

        return "%s %s %s" % (self.this, 
                             self.operator, 
                             self.that)

class SqlNullFilter(SqlComparisonFilter):
    def __init__(self, this):
        super(SqlNullFilter, self).__init__(this, 
                                            "null", 
                                            "is not",
                                            valid_operators=("is not",))

class SqlValueFilter(SqlComparisonFilter):
    def __init__(self, this, operator="=", valid_operators=("<",">","=","<=",">=")):
        that = "%%(%s)s" % this.name
        super(SqlValueFilter, self).__init__(this, that, operator, 
                                             valid_operators=valid_operators)

class SqlLikeFilter(SqlValueFilter):
    BEGINS = "B"
    CONTAINS = "C"
    def __init__(self, this, operator="like"):
        super(SqlLikeFilter, self).__init__(this, operator, valid_operators=("like",))

class SqlILikeFilter(SqlValueFilter):
    BEGINS = "B"
    CONTAINS = "C"
    def __init__(self, this, operator="ilike"):
        super(SqlILikeFilter, self).__init__(this, operator, valid_operators=("ilike",))

class SqlExistenceFilter(SqlOpFilter):
    def __init__(self, subquery, operator="exists"):
        super(SqlExistenceFilter, self).__init__(operator, "exists")

        self.subquery = subquery

    def emit(self):
        self.check_operator()
        return "%s (%s)" % (self.operator, self.subquery)

class SqlDateValueFilter(SqlOpFilter):
    def __init__(self, this, that, operator="="):
        super(SqlDateValueFilter, self).__init__(operator, "<",">","=","<=",">=")

        assert isinstance(operator, str)

        self.this = getattr(this, "identifier", this)
        self.that = getattr(that, "identifier", that)

    def emit(self):
        self.check_operator()
        return "%s %s date('%s')" % (self.this, self.operator, self.that)
